#include <stdio.h>

unsigned readbit(unsigned i, 
		 unsigned j,
		 unsigned bm[8]) {
  // i, j: 0 .. 15
  return ((bm[((i&15) >> 1)] >> (31 - (16*(i&1) + j))) & 1);
}

void writebit (unsigned bit,
               unsigned i, 
	       unsigned j,
	       unsigned bm[8]) {
  // i, j: 0 .. 15
  if (bit&1)
    bm[((i&15) >> 1)] |=  (1 << (31 - (16*(i&1) + j)));
  else
    bm[((i&15) >> 1)] &= ~(1 << (31 - (16*(i&1) + j)));
}

void bmm_sw(unsigned cm[8], 
	    unsigned am[8], 
	    unsigned bm[8]) {
  unsigned i, j, k;

  // This is the reference implementation of
  // BMM-16. This routine is used during performance
  // profiling and functional verification. You
  // cannot change anything in this function, nor in
  // writebit() or readbit()

  for (i=0; i<16; i++) {
    for (j=0; j<16; j++) {
      writebit(0, i, j, cm);
      for (k=0; k<16; k++) {
	writebit(readbit(i, j, cm) ^ 
		 (readbit(i, k, am) & 
		  readbit(k, j, bm)), i, j, cm);
      }
    }
  }
}

void bmm_hw(unsigned cm[8], 
	    unsigned am[8], 
	    unsigned bm[8]) {

  // This function is the entry point for your 
  // hardware driver. Replace the function call
  // to bmm_sw with a driver suitable to your
  // hardware coprocessor.

  bmm_sw(cm, am, bm);
}

//----------------------- testbench -------------------------
//
// Everything below this comment line has to remain unchanged!
   
void genmatrix(unsigned am[8]) {
  unsigned i, j, k;
  for (i=0; i<8; i++)
    am[i] = (random() << 1) | (random() & 1);
}

// shared data area
unsigned am[8];
unsigned bm[8];
unsigned sw_cm[8];
unsigned hw_cm[8];

#define FUNCLOOP 5
#define TIMELOOP 5

int main() {
  long long swtiming[TIMELOOP];
  long long hwtiming[TIMELOOP];
  long long avg_swtiming;
  long long avg_hwtiming;
  unsigned i, j, err;

  printf("Functional Verification\n");
  // functional verification
  for (i=0; i<FUNCLOOP; i++) {
    printf(".\n");
    genmatrix(am);
    genmatrix(bm);
    bmm_sw(sw_cm, am, bm);
    bmm_hw(hw_cm, am, bm);
    err = 0;
    for (j=0; j<8; j++) 
      if (sw_cm[j] != hw_cm[j]) 
	err = 1;
    if (err)  {
      printf("Difference detected for (A, B, SW-C, HW-C):\n");
      for (j=0; j<8; j++) 
	printf("%08x %08x %08x %08x\n", am[j], bm[j], sw_cm[j], hw_cm[j]);
      exit(0);
    }
  }
  printf("Functional Verification OK!\n");

  printf("Performance Timing\n");
  // sw timing
  for (i=0; i<TIMELOOP; i++) {
    genmatrix(am);
    genmatrix(bm);
    swtiming[i] = getcyclecount();
    bmm_sw(sw_cm, am, bm);
    swtiming[i] = getcyclecount() - swtiming[i];
  }
  
  // hw timing
  for (i=0; i<TIMELOOP; i++) {
    genmatrix(am);
    genmatrix(bm);
    hwtiming[i] = getcyclecount();
    bmm_hw(hw_cm, am, bm);
    hwtiming[i] = getcyclecount() - hwtiming[i];
  }
  
  avg_swtiming = 0;
  avg_hwtiming = 0;
  for (i=0; i<TIMELOOP; i++) {
    avg_swtiming += swtiming[i];
    avg_hwtiming += hwtiming[i];
  }  
  printf("Average SW cycles: %lld\n", avg_swtiming);
  printf("Average HW cycles: %lld\n", avg_hwtiming);
  printf("Speedup: %f\n", 1.0 * avg_swtiming / avg_hwtiming);

  return 0;
}
